.onload <- function(libname, pkgname)
{
    library.dynam("L0ggm",package=pkgname,lib.loc=libname)
}

.onAttach <- function(libname, pkgname)
{
    msg <- styletext(styletext(paste("\nL0ggm (version ", packageVersion("L0ggm"), ")", sep = ""), defaults = "underline"), defaults = "bold")
    # msg <- paste(msg,'\n\nFor help getting started, see <https://r-ega.net>')
    msg <- paste(msg,"\n\nFor bugs and errors, submit an issue to <https://github.com/AlexChristensen/L0ggm/issues>")

    packageStartupMessage(msg)
}
