#' Toy Small-world Network Data Example
#'
#' A toy small-world network example (\emph{n} = 500)
#'
#' @name basic_smallworld
#'
#' @docType data
#'
#' @usage data(basic_smallworld)
#'
#' @format A 500 \eqn{\times} 20 continuous response matrix
#'
#' @keywords datasets
#'
#' @examples
#' # Generated using:
#' # basic_smallworld <- simulate_smallworld(
#' #  nodes = 20, # 20 nodes in the network
#' #  density = 0.30, # moderate initial lattice connectivity
#' #  rewire = 0.20, # 20% rewiring probability
#' #  sample_size = 500 # number of cases = 500
#' # )$data
#' data("basic_smallworld")
#'
NULL
#----
