#' @title L0ggm-package
#'
#' @description Implements L0 norm regularization penalties for Gaussian graphical models
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @useDynLib L0ggm, .registration = TRUE
#'
#' @importFrom methods is
#' @importFrom stats complete.cases cor cov2cor dist dnorm na.omit optimize pchisq pnorm predict qnorm rnorm sd setNames uniroot
#' @importFrom utils capture.output data packageVersion
#'
"_PACKAGE"
#> [1] "_PACKAGE"
#L0ggm----
