#' Skew Tables
#'
#' Tables for skew based on the number of categories (2, 3, 4, 5, or 6) in the data
#'
#' @name skew_tables
#'
#' @docType data
#'
#' @usage data(skew_tables)
#'
#' @format A list (length = 5)
#'
#' @keywords datasets
#'
#' @examples
#' data("skew_tables")
#'
NULL
