% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/L0ggm.R
\docType{package}
\name{L0ggm-package}
\alias{L0ggm}
\alias{L0ggm-package}
\title{L0ggm-package}
\description{
Implements L0 norm regularization penalties for Gaussian graphical models
}
\seealso{
Useful links:
\itemize{
  \item Report bugs at \url{https://github.com/AlexChristensen/L0ggm/issues}
}

}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}
